/*
 * DLPI test program. See Usage().
 */

#include <sys/types.h>
#include <stropts.h>
#include <sys/conf.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/stream.h>
#include <sys/dlpi.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <inttypes.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>

/* option flags */
int aflag = 0;
t_uscalar_t ppa = 0;
int bflag = 0;
t_uscalar_t sap = 0x1010;
t_uscalar_t dlsap_addr_len;
int pflag_phys = 0;
int pflag_sap = 0;
int pflag_multi = 0;
int rflag = 0;
int reply_mode = 0;
int sflag = 0;

char *device = NULL;

/* data link addresses */
#define DLADDR_MAX 16
typedef unsigned char t_data_link_addr[DLADDR_MAX];
int data_link_addr_len;

/* dlsap addresses */
#define DLSAP_LEN	(DLADDR_MAX + sizeof(t_uscalar_t))
typedef unsigned char t_dlsap_addr[DLSAP_LEN];
int sap_length;  /* as returned by DL_INFO_ACK, but >= 0 */
enum sap_ordering { SO_UNDEF = 0, SO_SAP_PHYS, SO_PHYS_SAP };
enum sap_ordering sap_ordering = SO_UNDEF;

t_data_link_addr send_addr; 	/* for -s */
uchar_t *send_str = NULL; 		/* for -s */

int fd;	/* fd for dlpi device */

#define PRIMBUFMAX 512
uchar_t primbuf[PRIMBUFMAX];
#define BUFMAX (64*1024)
uchar_t buf[BUFMAX];

void xdump(unsigned char *p, int l);
void ether_header_dump(struct ether_header *p);
void nbytes_hex(char *s1, uchar_t *b, int n, char *s2);
void ascii_out(unsigned char *p, int n);
void dump_ether_packet(struct ether_header *p, int l);
void dump_ip_packet(struct ip *p, int l);
void dump_in_addr(char *s1, struct in_addr a, char *s2);
char *getipp(int prot);
char *get_ether_type(int ty);
void Usage(char *prog);
void Error(char *str, ...);
void DlpiSet(t_uscalar_t prim_type, void *prim, int len);
void PutDlpiMsg(void *prim, int prim_len, void *data, int data_len);
int GetDlpiMsg(void *prim, int prim_len, void *data, int *data_max);
void DlpiChk(t_uscalar_t prim_type, void *prim);
char *get_dl_mac_type(t_uscalar_t ty);
char *get_dl_service_mode(t_uscalar_t m);
void DlpiChkOk(t_uscalar_t correct_primitive);
void dlsap_to_dl_plus_sap(t_dlsap_addr dlsap, t_data_link_addr dl_addr, 
	t_uscalar_t *psap);
void dl_plus_sap_to_dlsap(t_data_link_addr dl_addr, 
		t_uscalar_t sap, t_dlsap_addr dlsap);
t_uscalar_t bytes_to_uscalar(uchar_t *p, int l);
void uscalar_to_bytes(uchar_t *p, int l, t_uscalar_t u);
void DlpiSnd(t_dlsap_addr dlsap_addr, uchar_t *buf, int len);
void DlpiRcv(t_dlsap_addr src_addr, t_dlsap_addr dest_addr, 
	uchar_t *buf, int *len);
void DlpiPromiscon(t_uscalar_t promisc_type, char *s);

int
main(int argc, char **argv)
{
	int c, n;
	pid_t pid;
	char *p;
	int send_addr_len = 0, len;
	unsigned h;
	t_uscalar_t style;
	t_uscalar_t service_mode;
	dl_info_req_t dl_info_req;
	dl_info_ack_t *p_info_ack;
	dl_phys_addr_req_t dl_phys_addr_req;
	dl_attach_req_t dl_attach_req;
	dl_phys_addr_ack_t *p_phys_addr_ack;
	dl_bind_req_t dl_bind_req;
	dl_bind_ack_t *p_bind_ack;
	t_data_link_addr dl_addr;
	t_uscalar_t sap2;
	t_dlsap_addr dlsap_addr, dest_addr, my_addr;

	setbuf(stdout, NULL);
	setbuf(stderr, NULL);

	while ((c = getopt(argc, argv, "a:b:p:rRs")) != EOF) {
	    switch(c) {
		case 'a':
			aflag = 1;
			ppa = strtoul(optarg, NULL, 0);
			break;
		case 'b':
			bflag = 1;
			sap = strtoul(optarg, NULL, 0);
			break;
		case 'p':
			if (!strcmp(optarg,"p") || !strcmp(optarg,"phys"))
				pflag_phys = 1;
			else if (!strcmp(optarg,"s") || !strcmp(optarg,"sap"))
				pflag_sap = 1;
			else if (!strcmp(optarg,"m") || !strcmp(optarg,"multi"))
				pflag_multi = 1;
			else
				Error("-p option: unknown level %s", optarg);
			break;
		case 'r':
			rflag = 1;
			reply_mode = 0;
			break;
		case 'R':
			rflag = 1;
			reply_mode = 1;
			break;
		case 's':
			sflag = 1;
			break;
		case '?':
		default:
			Usage(argv[0]);
	    }
	}

	if (!sflag) {
		if (argc != optind + 1)
			Usage(argv[0]);
	}
	else {
		if (argc != optind + 3)
			Usage(argv[0]);

		send_addr_len = 0;
		p = argv[optind+1];
		while (*p) {
			if (send_addr_len >= DLADDR_MAX)
				Error("-s option: max supported addr len %d",
					DLADDR_MAX);

			if (sscanf(p, "%02x%n", &h, &n) 
					!= 1 || (p[n] && p[n] != ':'))
				Error("-s option: invalid send address %s",
					argv[optind+1]);
			send_addr[send_addr_len] = h;
			p += n;
			if (*p)
				++p;
			++send_addr_len;
		}

		send_str = (uchar_t *)argv[optind+2];
		nbytes_hex("sending to ", send_addr, send_addr_len, "\n");
	}

	if (sap <= ETHERMTU + sizeof(struct ether_header))
		printf("Warning: sap 0x%x < 0x%x (%d), IEEE 802.3 mode?\n", 
			sap, ETHERMTU + sizeof(struct ether_header),
			ETHERMTU + sizeof(struct ether_header));

	device = argv[optind];

	fd = open(device, O_RDWR);
	if (fd < 0)
		Error("cannot open device %s", device);
	
	/*** DL_INFO_REQ **************************************************/
	DlpiSet(DL_INFO_REQ, &dl_info_req, sizeof(dl_info_req));
	PutDlpiMsg(&dl_info_req, sizeof(dl_info_req), NULL, 0);
	p_info_ack = (dl_info_ack_t *)primbuf;
	GetDlpiMsg(p_info_ack, PRIMBUFMAX, NULL, NULL);
	DlpiChk(DL_INFO_ACK, p_info_ack);
	printf("dl_max_sdu %lu, dl_min _sdu %lu\n", p_info_ack->dl_max_sdu,
		p_info_ack->dl_min_sdu);
	printf("dl_addr_length %lu\n", p_info_ack->dl_addr_length);
	printf("dl_addr_offset %lu\n", p_info_ack->dl_addr_offset);
	printf("dl_mac_type %lu = %s\n", p_info_ack->dl_mac_type, 
		get_dl_mac_type(p_info_ack->dl_mac_type));
	printf("dl_sap_length %ld\n", p_info_ack->dl_sap_length);
	printf("dl_service_mode %lu = %s\n", p_info_ack->dl_service_mode,
		get_dl_service_mode(p_info_ack->dl_service_mode));
	printf("dl_provider_style 0x%lx", p_info_ack->dl_provider_style);
	switch (p_info_ack->dl_provider_style) {
	    case DL_STYLE1: printf(" = DL_STYLE1\n"); break;
	    case DL_STYLE2: printf(" = DL_STYLE2\n"); break;
	    default: printf(" ???\n");
	}
	printf("dl_version %lu\n", p_info_ack->dl_version);
	style = p_info_ack->dl_provider_style;
	service_mode = p_info_ack->dl_service_mode;

	/*** DL_ATTACH_REQ ************************************************/
	if (style == DL_STYLE2 || aflag) {
		printf("attaching to ppa %d\n", ppa);
		DlpiSet(DL_ATTACH_REQ, &dl_attach_req, sizeof(dl_attach_req_t));
		dl_attach_req.dl_ppa = ppa;
		PutDlpiMsg(&dl_attach_req, sizeof(dl_attach_req_t), NULL, 0);
		DlpiChkOk(DL_ATTACH_REQ);
	}

	/*** DL_PHYS_ADDR_REQ *********************************************/
	DlpiSet(DL_PHYS_ADDR_REQ, &dl_phys_addr_req, sizeof(dl_info_req));
	dl_phys_addr_req.dl_addr_type = DL_CURR_PHYS_ADDR;
	PutDlpiMsg(&dl_phys_addr_req, sizeof(dl_phys_addr_req), NULL, 0);
	p_phys_addr_ack = (dl_phys_addr_ack_t *)primbuf;
	GetDlpiMsg(p_phys_addr_ack, PRIMBUFMAX, NULL, NULL);
	DlpiChk(DL_PHYS_ADDR_ACK, p_phys_addr_ack);
	nbytes_hex("current pys addr ", 
		(uchar_t *)p_phys_addr_ack + p_phys_addr_ack->dl_addr_offset, 
		p_phys_addr_ack->dl_addr_length, "\n");
	data_link_addr_len = p_phys_addr_ack->dl_addr_length;
	if (sflag && send_addr_len != data_link_addr_len)
		Error("-s option: invalid address length %d, should be %d",
			send_addr_len, data_link_addr_len);

	/*** DL_BIND_REQ **************************************************/
	printf("binding to sap 0x%x\n", sap);
	DlpiSet(DL_BIND_REQ, &dl_bind_req, sizeof(dl_bind_req));
	dl_bind_req.dl_sap = sap;
	dl_bind_req.dl_max_conind = 0;
	dl_bind_req.dl_service_mode = service_mode;
	dl_bind_req.dl_conn_mgmt = 0;
	dl_bind_req.dl_xidtest_flg = 0;
	PutDlpiMsg(&dl_bind_req, sizeof(dl_bind_req), NULL, 0);
	p_bind_ack = (dl_bind_ack_t *)primbuf;
	GetDlpiMsg(p_bind_ack, PRIMBUFMAX, NULL, NULL);
	DlpiChk(DL_BIND_ACK, p_bind_ack);
	nbytes_hex("bind address ", 
		(uchar_t *)p_bind_ack + p_bind_ack->dl_addr_offset, 
		p_bind_ack->dl_addr_length, "\n");

	/*** DL_PROMISCON_REQ *********************************************/
	if (pflag_phys)
		DlpiPromiscon(DL_PROMISC_PHYS, "DL_PROMISC_PHYS");
	if (pflag_sap)
		DlpiPromiscon(DL_PROMISC_SAP, "DL_PROMISC_SAP");
	if (pflag_multi)
		DlpiPromiscon(DL_PROMISC_MULTI, "DL_PROMISC_MULTI");

	/*** second DL_INFO_REQ, after bind, get dlsap addr info **********/
	DlpiSet(DL_INFO_REQ, &dl_info_req, sizeof(dl_info_req));
	PutDlpiMsg(&dl_info_req, sizeof(dl_info_req), NULL, 0);
	p_info_ack = (dl_info_ack_t *)primbuf;
	GetDlpiMsg(p_info_ack, PRIMBUFMAX, NULL, NULL);
	if (p_info_ack->dl_addr_offset <= 0)
		Error("no dlsap address after bind");
	DlpiChk(DL_INFO_ACK, p_info_ack);
	printf("dl_sap_length %ld\n", p_info_ack->dl_sap_length);
	if (p_info_ack->dl_sap_length == 0)
		Error("dl_sap_length = 0 after bind");
	if (p_info_ack->dl_sap_length > 0) {
		sap_length = p_info_ack->dl_sap_length;
		sap_ordering = SO_SAP_PHYS;
	}
	else {
		sap_length = - p_info_ack->dl_sap_length;
		sap_ordering = SO_PHYS_SAP;
	}
	dlsap_to_dl_plus_sap((uchar_t *)p_info_ack + 
		p_info_ack->dl_addr_offset, dl_addr, &sap2);
	if (data_link_addr_len + sap_length != p_info_ack->dl_addr_length)
		Error("data_link_addr_len (%d) + sap_length (%d) "
			"!= dl_addr_length (%d)", data_link_addr_len,
			sap_length, p_info_ack->dl_addr_length);
	dlsap_addr_len = p_info_ack->dl_addr_length;
	memcpy(my_addr, (uchar_t *)p_info_ack + p_info_ack->dl_addr_offset, 
		dlsap_addr_len);
	nbytes_hex("full dlsap address ", my_addr,
		p_info_ack->dl_addr_length, " = \n");
	nbytes_hex("   dll addr ", dl_addr, data_link_addr_len, ", ");
	printf("sap 0x%x\n", sap2);

	if (sflag) {
		dl_plus_sap_to_dlsap(send_addr, sap, dlsap_addr);
		nbytes_hex("sending to ", dlsap_addr, dlsap_addr_len, "\n");
		DlpiSnd(dlsap_addr, send_str, strlen((char *)send_str));
	}

	pid = getpid();
	n = 1;

	while (rflag) {
		len = BUFMAX - 1;
		printf("receiving ...\n");
		DlpiRcv(dlsap_addr, dest_addr, buf, &len);

		if (reply_mode && !memcmp(dest_addr, my_addr, dlsap_addr_len)) {
			sprintf((char *)buf, "reply %d from pid %d", n++, pid);
			nbytes_hex("replying to ", dlsap_addr, 
				dlsap_addr_len, "\n");
			DlpiSnd(dlsap_addr, buf, strlen((char *)buf));
		}
	}
}

void
Usage(char *prog)
{
   fprintf(stderr, "usage: %s [options] device\n", prog);
   fprintf(stderr, "       %s -s [options] device data_link_addr string\n", 
		prog);
   fprintf(stderr, "       -a ppa:        attach to ppa "
					  "(for style2, default 0)\n");
   fprintf(stderr, "       -b sap:        bind to sap (default 0x%X)\n", sap);
   fprintf(stderr, "       -p level:      promiscuous mode (repeated ...)\n");
   fprintf(stderr, "                      level = p[hys] s[ap] m[ulti]\n");
   fprintf(stderr, "       -r:            receive data\n");
   fprintf(stderr, "       -R:            receive data and reply\n");
   fprintf(stderr, "       -s:            send string to data_link_addr\n");
   fprintf(stderr, "                      data_link_addr xx:xx:xx:...\n");

   exit(1);
}

void
xdump(unsigned char *p, int l)
{
	int n;

	for (n = 0; n < l; ++n) {
		if (!(n & 0xf))
			printf("    %04x: ", n);
		if (!(n & 0x7))
			putchar(' ');
		printf("%02x ", p[n]);
		if ((n & 0xf) == 0xf)
			ascii_out(p, n);
	}

	if (n & 0xf)
		ascii_out(p, n-1);
}

void
ascii_out(unsigned char *p, int n)
{
	int i;

	for (i = (n & 0xf) + 1; i < 16; ++i) {
		if (i == 8)
			putchar(' ');
		printf("   ");
	}
	printf(" ");
	for (i = n & ~0xf; i <= n; ++i) {
		if (!(i & 0x7))
			putchar(' ');
		if (isprint(p[i]))
			putchar(p[i]);
		else
			putchar('.');
	}
	putchar('\n');
}

void
ether_header_dump(struct ether_header *p)
{
	int t;

	nbytes_hex("    ether_dhost ",p->ether_dhost.ether_addr_octet, 6, "\n");
	nbytes_hex("    ether_shost ",p->ether_shost.ether_addr_octet, 6, "\n");
	t = ntohs(p->ether_type);
	printf("    ether_type 0x%x = %s\n", t, get_ether_type(t));
}

void 
dump_ether_packet(struct ether_header *p, int l)
{
	int t;

	printf("  -- ETHER --\n");
	if (l < sizeof(struct ether_header))
		return;
	ether_header_dump(p);
	t = ntohs(p->ether_type);
	switch (t) {
	    case ETHERTYPE_IP:
		dump_ip_packet((struct ip *)(p + 1), 
			l - sizeof(struct ether_header));
		break;
	}
		
}

void
dump_ip_packet(struct ip *p, int l)
{
	printf("  -- IP --\n");
	if (l < sizeof(struct ip))
		return;
	dump_in_addr("    ip_src ", p->ip_src, "\n");
	dump_in_addr("    ip_dst ", p->ip_dst, "\n");
	printf("    ip_p %d = %s\n", p->ip_p, getipp( p->ip_p));
}

void
dump_in_addr(char *s1, struct in_addr a, char *s2)
{
	if (s1)
		printf(s1);
	printf("%d.%d.%d.%d", a._S_un._S_un_b.s_b1, a._S_un._S_un_b.s_b2,
		a._S_un._S_un_b.s_b3, a._S_un._S_un_b.s_b4);
	if (s2)
		printf(s2);
}

void 
nbytes_hex(char *s1, uchar_t *b, int n, char *s2)
{
	int k;

	if (s1)
		printf(s1);
	printf("%02x", b[0]);
	for (k = 1; k < n; ++k)
		printf(":%02x", b[k]);
	if (s2)
		printf(s2);
}

void
Error(char *str, ...)
{
        va_list ap;
	int e;

	e = errno;

        va_start(ap, str);
        vfprintf(stderr, str, ap);
	va_end(ap);
        fprintf( stderr, "\n" );

	if (e)
		fprintf(stderr, "system error %d: %s\n", e, strerror(e));

        exit(1);
}

char *
getipp(int prot)
{
	switch (prot) {
	    case IPPROTO_HOPOPTS:	return "IPPROTO_HOPOPTS";
	    case IPPROTO_ICMP:		return "IPPROTO_ICMP";
	    case IPPROTO_IGMP:		return "IPPROTO_IGMP";
	    case IPPROTO_GGP:		return "IPPROTO_GGP";
	    case IPPROTO_ENCAP:		return "IPPROTO_ENCAP";
	    case IPPROTO_TCP:		return "IPPROTO_TCP";
	    case IPPROTO_EGP:		return "IPPROTO_EGP";
	    case IPPROTO_PUP:		return "IPPROTO_PUP";
	    case IPPROTO_UDP:		return "IPPROTO_UDP";
	    case IPPROTO_IDP:		return "IPPROTO_IDP";
	    case IPPROTO_IPV6:		return "IPPROTO_IPV6";
	    case IPPROTO_ROUTING:	return "IPPROTO_ROUTING";
	    case IPPROTO_FRAGMENT:	return "IPPROTO_FRAGMENT";
	    case IPPROTO_RSVP:		return "IPPROTO_RSVP";
	    case IPPROTO_ESP:		return "IPPROTO_ESP";
	    case IPPROTO_AH:		return "IPPROTO_AH";
	    case IPPROTO_ICMPV6:	return "IPPROTO_ICMPV6";
	    case IPPROTO_NONE:		return "IPPROTO_NONE";
	    case IPPROTO_DSTOPTS:	return "IPPROTO_DSTOPTS";
	    case IPPROTO_HELLO:		return "IPPROTO_HELLO";
	    case IPPROTO_ND:		return "IPPROTO_ND";
	    case IPPROTO_EON:		return "IPPROTO_EON";
	    case IPPROTO_PIM:		return "IPPROTO_PIM";
	    case IPPROTO_RAW:		return "IPPROTO_RAW";
	    default: return "???";
	}
}

char *
get_ether_type(int ty)
{
	switch (ty) {
	    case 0:			return "ALL";
	    case ETHERTYPE_PUP:		return "ETHERTYPE_PUP";
	    case ETHERTYPE_IP:		return "ETHERTYPE_IP";
	    case ETHERTYPE_ARP:		return "ETHERTYPE_ARP";
	    case ETHERTYPE_REVARP:	return "ETHERTYPE_REVARP";
	    case ETHERTYPE_IPV6:	return "ETHERTYPE_IPV6";
	    default: 			return "???";
	}
}

void
DlpiSet(t_uscalar_t prim_type, void *prim, int len)
{
	memset(prim, 0, len);
	((union DL_primitives *)prim)->dl_primitive = prim_type;
}

void
PutDlpiMsg(void *prim, int prim_len, void *data, int data_len)
{
	int flags, e;
	struct strbuf ctlbuf, databuf;

	ctlbuf.buf = (char *)prim;
	ctlbuf.len = prim_len;

	databuf.buf = (char *)data;
	databuf.len = data_len;

	flags = 0;

	e = putmsg(fd, &ctlbuf, (data ? &databuf : NULL), flags);
	if (e)
		Error("putmsg prim 0x%x", 
			((union DL_primitives *)prim)->dl_primitive);
}

/* returns data length read
 */
int
GetDlpiMsg(void *prim, int prim_len, void *data, int *data_len)
{
	int flags, e;
	struct strbuf ctlbuf, databuf;

	ctlbuf.buf = (char *)prim;
	ctlbuf.maxlen = prim_len;

	databuf.buf = (char *)data;
	databuf.maxlen = (data_len ? *data_len : 0);

	flags = 0;

	e = getmsg(fd, &ctlbuf, &databuf, &flags);
	if (e & (MORECTL|MOREDATA))
		Error("getmsg MORECTL|MOREDATA= 0x%x, ctlbuf.len %d", 
			e, ctlbuf.len);
	if (e)
		Error("getmsg");
	
	if (data_len)
		*data_len = databuf.len;
}

void
DlpiChk(t_uscalar_t prim_type, void *prim)
{
	union DL_primitives *p = prim;

	if (p->dl_primitive == prim_type)
		return;

	if (p->dl_primitive == DL_ERROR_ACK)
		Error("DlpiChk: expected dl_primitive 0x%0x,\n"
			"         got DL_ERROR_ACK"
			" dl_errno 0x%x, dl_unix_errno %d",
			prim_type, p->error_ack.dl_errno, 
			p->error_ack.dl_unix_errno);

	Error("DlpiChk: expected dl_primitive 0x%0x, got 0x%0x",
		prim_type, p->dl_primitive);
}

char *
get_dl_mac_type(t_uscalar_t ty)
{
	switch (ty) {
	    case DL_CSMACD:	return "DL_CSMACD";
	    case DL_TPB:	return "DL_TPB";
	    case DL_TPR:	return "DL_TPR";
	    case DL_METRO: 	return "DL_METRO";
	    case DL_ETHER: 	return "DL_ETHER";
	    case DL_HDLC: 	return "DL_HDLC";
	    case DL_CHAR: 	return "DL_CHAR";
	    case DL_CTCA: 	return "DL_CTCA";
	    case DL_FDDI: 	return "DL_FDDI";
	    case DL_FC: 	return "DL_FC";
	    case DL_ATM: 	return "DL_ATM";
	    case DL_IPATM: 	return "DL_IPATM";
	    case DL_X25: 	return "DL_X25";
	    case DL_ISDN: 	return "DL_ISDN";
	    case DL_HIPPI: 	return "DL_HIPPI";
	    case DL_100VG: 	return "DL_100VG";
	    case DL_100VGTPR: 	return "DL_100VGTPR";
	    case DL_ETH_CSMA: 	return "DL_ETH_CSMA";
	    case DL_100BT: 	return "DL_100BT";
	    case DL_FRAME: 	return "DL_FRAME";
	    case DL_MPFRAME: 	return "DL_MPFRAME";
	    case DL_ASYNC: 	return "DL_ASYNC";
	    case DL_IPX25: 	return "DL_IPX25";
	    case DL_LOOP: 	return "DL_LOOP";
	    case DL_OTHER: 	return "DL_OTHER";
	    default: return "???";
	}
}

char *
get_dl_service_mode(t_uscalar_t m)
{
	switch (m) {
	    case DL_CODLS:	return "DL_CODLS";
	    case DL_CLDLS:	return "DL_CLDLS";
	    case DL_ACLDLS:	return "DL_ACLDLS";
	    default: return "???";
	}
}

void
DlpiChkOk(t_uscalar_t correct_primitive)
{
	dl_ok_ack_t *ok = (dl_ok_ack_t *)primbuf;
	union DL_primitives *p = (union DL_primitives *)primbuf;

	GetDlpiMsg(ok, PRIMBUFMAX, NULL, NULL);
	if (ok->dl_primitive == DL_OK_ACK && 
			ok->dl_correct_primitive == correct_primitive)
		return;

	if (ok->dl_primitive == DL_ERROR_ACK)
		Error(  "DlpiChkOk: expected dl_primitive DL_OK_ACK,\n"
			"           got DL_ERROR_ACK"
			" dl_errno 0x%x, dl_unix_errno %d",
			p->error_ack.dl_errno, 
			p->error_ack.dl_unix_errno);

	if (ok->dl_primitive != DL_OK_ACK)
		Error("DlpiChkOk: expected DL_OK_ACK got 0x%x", 
			ok->dl_primitive);
	if (ok->dl_correct_primitive != correct_primitive)
		Error("DlpiChkOk: expected correct_primitive 0x%x got 0x%x",
			correct_primitive, ok->dl_correct_primitive);
}

void 
dlsap_to_dl_plus_sap(t_dlsap_addr dlsap, t_data_link_addr dl_addr, 
	t_uscalar_t *psap)
{
	uchar_t *p_dl_addr;
	uchar_t *p_sap;

	if (sap_ordering == SO_UNDEF)
		Error("dlsap_to_dl_plus_sap: sap_ordering == SO_UNDEF");
	
	if (sap_ordering == SO_SAP_PHYS) {
		p_dl_addr = dlsap + sap_length;
		p_sap = dlsap;
	}
	else {
		p_dl_addr = dlsap;
		p_sap = dlsap + data_link_addr_len;
	}

	memcpy(dl_addr, p_dl_addr, data_link_addr_len);
	*psap = bytes_to_uscalar(p_sap, sap_length);
}

void 
dl_plus_sap_to_dlsap(t_data_link_addr dl_addr, 
		t_uscalar_t sap, t_dlsap_addr dlsap)
{
	uchar_t *p_dl_addr;
	uchar_t *p_sap;

	if (sap_ordering == SO_UNDEF)
		Error("dl_plus_sap_to_dlsap: sap_ordering == SO_UNDEF");
	
	if (sap_ordering == SO_SAP_PHYS) {
		p_dl_addr = dlsap + sap_length;
		p_sap = dlsap;
	}
	else {
		p_dl_addr = dlsap;
		p_sap = dlsap + data_link_addr_len;
	}

	memcpy(p_dl_addr, dl_addr, data_link_addr_len);
	uscalar_to_bytes(p_sap, sap_length, sap);
}

t_uscalar_t
bytes_to_uscalar(uchar_t *p, int l)
{
	int inc, n;
	t_uscalar_t res;

	if (l > sizeof(t_uscalar_t))
		Error("bytes_to_uscalar: l = %d", l);

#ifdef __i386
	inc = -1;
	p += l - 1;
#else
	inc = 1;
#endif

	res = 0;
	for (n = 0; n < l; ++n, p += inc)
		res = (res << 8) | *p;
	
	return res;
}

void
uscalar_to_bytes(uchar_t *p, int l, t_uscalar_t u)
{
	int inc, n;

	if (l > sizeof(t_uscalar_t))
		Error("uscalar_to_bytes: l = %d", l);

#ifdef __i386
	inc = 1;
#else
	inc = -1;
	p += l - 1;
#endif

	for (n = 0; n < l; ++n, p += inc) {
		*p = u & 0xff;
		u >>= 8;
	}
}

void DlpiSnd(t_dlsap_addr dlsap_addr, uchar_t *buf, int len)
{
	dl_unitdata_req_t *p_unitdata_req;

	p_unitdata_req = (dl_unitdata_req_t *)primbuf;
	DlpiSet(DL_UNITDATA_REQ, p_unitdata_req, sizeof(dl_unitdata_req_t));
	p_unitdata_req->dl_dest_addr_length = dlsap_addr_len;
	p_unitdata_req->dl_dest_addr_offset = sizeof(dl_unitdata_req_t);
	p_unitdata_req->dl_priority.dl_min = 0;
	p_unitdata_req->dl_priority.dl_max = 0;
	memcpy((uchar_t *)p_unitdata_req + p_unitdata_req->dl_dest_addr_offset,
		dlsap_addr, dlsap_addr_len);
	PutDlpiMsg(p_unitdata_req, sizeof(dl_unitdata_req_t) + dlsap_addr_len, 
		buf, len);
}

void DlpiRcv(t_dlsap_addr src_addr, t_dlsap_addr dest_addr, 
					uchar_t *buf, int *len)
{
	dl_unitdata_ind_t *p_unitdata_ind;
	uchar_t *p;

	p_unitdata_ind = (dl_unitdata_ind_t *)primbuf;
	p = (uchar_t *)p_unitdata_ind;
	GetDlpiMsg(p_unitdata_ind, PRIMBUFMAX, buf, len);
	DlpiChk(DL_UNITDATA_IND, p_unitdata_ind);

	printf("received:");
	if (p_unitdata_ind->dl_src_addr_offset > 0) {
		if (p_unitdata_ind->dl_src_addr_length != dlsap_addr_len)
			Error("DlpiRcv src_addr_length %d", 
				p_unitdata_ind->dl_src_addr_length);
		memcpy(src_addr, p + p_unitdata_ind->dl_src_addr_offset, 
			dlsap_addr_len);
		nbytes_hex(" src ", src_addr, dlsap_addr_len, NULL);
	}
	else
		memset(src_addr, 0, dlsap_addr_len);

	if (p_unitdata_ind->dl_dest_addr_offset > 0) {
		if (p_unitdata_ind->dl_dest_addr_length != dlsap_addr_len)
			Error("DlpiRcv dest_addr_length %d", 
				p_unitdata_ind->dl_dest_addr_length);
		memcpy(dest_addr, p + p_unitdata_ind->dl_dest_addr_offset, 
			dlsap_addr_len);
		nbytes_hex(" dest ", dest_addr, dlsap_addr_len, NULL);
	}
	else
		memset(dest_addr, 0, dlsap_addr_len);

	if (p_unitdata_ind->dl_group_address)
		printf(" group_address");
	printf(" data-len %d\n", *len);

	xdump(buf, *len);
}

void 
DlpiPromiscon(t_uscalar_t promisc_type, char *s)
{
	dl_promiscon_req_t promiscon_req;

	printf("setting promiscuous mode %s\n", s);
	DlpiSet(DL_PROMISCON_REQ, &promiscon_req, sizeof(dl_promiscon_req_t));
	promiscon_req.dl_level = promisc_type;
	PutDlpiMsg(&promiscon_req, sizeof(dl_promiscon_req_t), NULL, 0);
	DlpiChkOk(DL_PROMISCON_REQ);
}
