/*
 * usage: pts [command ...]
 *
 * Creates a master / slave pseudo tty pair and starts command or a shell
 * on the slave side. The master side's stdin is sent to the slave and the
 * slave's stdout and stderr are written to the master's stdout.
 *
 * This can be useful for running commands like telnet (Password:) that
 * refuse to read from a non tty input with stdin from a file etc.
 */

#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stropts.h>
#include <sys/conf.h>
#include <poll.h>

#define DELAY 500

void Error(char *str, ...);
void DoChild(int argc, char **argv);

int mfd;	/* master fd */

int
main(int argc, char **argv)
{
	pid_t pid;
	struct pollfd pfd;
	char c;
	int n, e, timeout;

	mfd = open("/dev/ptmx", O_RDWR);

	if (mfd < 0)
		Error("cannot open /dev/ptmx");
	grantpt(mfd);
	unlockpt(mfd);

	pid = fork();
	if (pid < 0)
		Error("cannot fork");
	
	if (!pid) {
		DoChild(argc, argv);
		exit(0);
	}

	pfd.fd = mfd;
	pfd.events = POLLIN|POLLOUT;

	poll(0, 0, DELAY);
	timeout = 0;
	for (;;) {
		if ((e = poll(&pfd, 1, timeout)) < 0)
			Error("poll %d", e);
		if (!e) {
			close(mfd);
			break;
		}

		if (pfd.revents & POLLIN) {
			n = read(mfd, &c, 1);
			if (n < 0)
				Error("read mfd");
			if (!n)
				break;
			write(1, &c, 1);
		}
		else if (pfd.revents & POLLOUT) {
			n = read(0, &c, 1);
			if (n < 0)
				Error("read 0");
			if (!n) {
				pfd.events = POLLIN;
				timeout = 200;
			}
			else {
				if (write(mfd, &c, 1) != 1)
					Error("write mfd");
				if (c == '\n')
					poll(0, 0, DELAY);
			}
		}
	}

	while (waitpid(pid, NULL, 0) != pid)
		;
	printf("\n");
	
	exit(0);
}

void
DoChild(int argc, char **argv)
{
	char *sh, *slavename;
	int fd;

	slavename = ptsname(mfd); 
	if (!slavename)
		Error("ptsname");
	close(mfd);

	if (setsid() < 0)
		Error("setsid failed");

	close(0);
	close(1);
	close(2);

	fd = open(slavename, O_RDWR); /* will become new control terminal */
	if (fd != 0)
		exit(1); /* no message possible */

	if (dup2(fd, 1) < 0)
		Error("dup 1");
	if (dup2(fd, 2) < 0)
		Error("dup 2");

	if (ioctl(fd, I_PUSH, "ptem") < 0)
		Error("I_PUSH ptem");
	if (ioctl(fd, I_PUSH, "ldterm") < 0)
		Error("I_PUSH ldterm");

	if ( argc < 2 ) {
		if ( !(sh=getenv("SHELL")) )
			sh="sh";

		execlp( sh, sh, NULL );
		Error("cannot execute %s\n", sh);
	}
	else
	{
		++argv;
		execvp(argv[0], argv);
		Error("cannot execute %s\n", argv[0]);

	}
}

void
Error(char *str, ...)
{
	int e;
        va_list ap;

	e = errno;

        va_start(ap, str);
        vfprintf(stderr, str, ap);
	va_end(ap);
        fprintf( stderr, "\n" );
	if (e)
		fprintf(stderr, "system error %d: %s\n", e, strerror(e));


        exit(1);
}

