/*
 * Uses /dev/sspy to trace streams messages.
 *
 * usage: xtrace [-a | -h]
 *
 *	  -a: Dump all message contents.
 *	  -h: Help.
 */

#include <sys/types.h>
#include <stropts.h>
#include <sys/conf.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/stream.h>
#include <sys/dlpi.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <inttypes.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "sspy.h"

int aflag = 0;	/* -a option */
#define MAXDATA (64*1024)

void Error(char *str, ...);
char *gettype(int t);
char *getop(int t);
void xdump(unsigned char *p, int l);
void (*dlpidump(void *p, int ty, int l)) 
	(void *p, int ty, int l, t_uscalar_t dl_sap);
void ioc_hdr_info_handler(void *p, int ty, int l, t_uscalar_t dl_sap);
void unitdata_req_dump(dl_unitdata_req_t *p);
void ether_header_dump(struct ether_header *p);
void nbytes_hex(char *s1, uchar_t *b, int n, char *s2);
void ascii_out(unsigned char *p, int n);
void dump_instance_sap(void *p, int ty, int l, t_uscalar_t dl_sap);
void dump_ether_packet(struct ether_header *p, int l);
void dump_ip_packet(struct ip *p, int l);
void dump_in_addr(char *s1, struct in_addr a, char *s2);
char *getipp(int prot);
char *get_dl_mac_type(t_uscalar_t ty);
char *get_dl_service_mode(t_uscalar_t m);
char *get_ether_type(int ty);
char *get_dl_level(int l);

int
main(int argc, char **argv)
{
	int fd, flag, e, l, ty, ty1, data_cnt;
	struct strbuf ctlbuf, databuf;
	struct sspy_trace_ctl ctl;
	unsigned char unaligned_data[MAXDATA+7], *p, *p2;
	unsigned char *data;
	t_uscalar_t dl_sap;
	struct ether_header *peh;
	void (*handler)(void *p, int ty, int l, t_uscalar_t dl_sap);

	setbuf(stdout, NULL);
	setbuf(stderr, NULL);

	if (argc > 1) {
		if (!strcmp(argv[1], "-h"))
			Error("usage: %s [-a | -h]", argv[0]);
		aflag = 1;
	}

	fd = open(SSPYDEV, O_RDWR);
	if (fd < 0)
		Error("cannot open %s", SSPYDEV);
	
	data = SSP_ALIGN8(unaligned_data);
	for (;;) {
		ctlbuf.buf = (char *)&ctl;
		ctlbuf.maxlen = sizeof(struct sspy_trace_ctl);
		databuf.buf = (char *)data;
		databuf.maxlen = MAXDATA;
		flag = 0;

		e = getmsg(fd, &ctlbuf, &databuf, &flag);
		if (e & (MORECTL|MOREDATA))
			Error("getmsg MORECTL|MOREDATA= 0x%x", e);
		if (e)
			Error("getmsg");
		
		dl_sap = ctl.dlpi_info.dl_sap;
		printf("%s, i %d, seq %u, e %d, l %d, ", getop(ctl.op),
			ctl.inst, ctl.seqno, ctl.err, databuf.len);
		if (ctl.dlpi_info.addr_len == SSPY_DLPI_UNBOUND)
			printf("unbound");
		else {
			printf("bound to sap 0x%x ", dl_sap);
			nbytes_hex(", addr ", ctl.dlpi_info.addr, 
				ctl.dlpi_info.addr_len, NULL);
		}
		printf("\n");

		if (databuf.len < 0)
			continue;

		handler = NULL;
		data_cnt = 0;
		for (p = data; p < data + databuf.len; p = SSP_MSG_NEXT(p)) {
			l = SSP_MSG_LEN(p);
			ty = SSP_MSG_TYPE(p);
			printf("  %s, len %d\n", gettype(ty), l);

			p2 = (unsigned char *)SSP_MSG_START(p);
			peh = (struct ether_header *)p2;

			if (ty == M_DATA)
				++data_cnt;
			if (p == data)
				ty1 = ty;

			if (p == data && ty == M_DATA) {
				if (ctl.op == SSPY_OP_WPUT)
				    dump_ether_packet(peh, l);
				else if (ctl.op == SSPY_OP_RPUT)
				    if (ctl.dlpi_info.flags&SSPY_DLPI_DLIOCRAW)
					dump_ether_packet(peh, l);
				    else
					dump_instance_sap(p2, ty, l, dl_sap);
			}
			else if (p == data && ty != M_DATA)
				handler = dlpidump(p2, ty, l);
			else if (handler)
				(*handler)(p2, ty, l, ctl.dlpi_info.dl_sap);
			else if (ty == M_DATA && ty1 == M_PROTO)
				dump_instance_sap(p2, ty, l, dl_sap);

			if (aflag)
				xdump(p2, l);
		}
	}
		
}

void
xdump(unsigned char *p, int l)
{
	int n;

	for (n = 0; n < l; ++n) {
		if (!(n & 0xf))
			printf("    %04x: ", n);
		if (!(n & 0x7))
			putchar(' ');
		printf("%02x ", p[n]);
		if ((n & 0xf) == 0xf)
			ascii_out(p, n);
	}

	if (n & 0xf)
		ascii_out(p, n-1);
}

void
ascii_out(unsigned char *p, int n)
{
	int i;

	for (i = (n & 0xf) + 1; i < 16; ++i) {
		if (i == 8)
			putchar(' ');
		printf("   ");
	}
	printf(" ");
	for (i = n & ~0xf; i <= n; ++i) {
		if (!(i & 0x7))
			putchar(' ');
		if (isprint(p[i]))
			putchar(p[i]);
		else
			putchar('.');
	}
	putchar('\n');
}

void (*dlpidump(void *p, int ty, int l)) 
		(void *p, int ty, int l, t_uscalar_t dl_sap)
{
	t_uscalar_t prim;
	void (*h)(void *p, int ty, int l, t_uscalar_t dl_sap) = NULL;

	if (ty == M_IOCTL || ty == M_IOCACK || ty == M_IOCNAK) {
		struct iocblk *i = p;

		printf("    ioc_cmd 0x%x", i->ioc_cmd);
		if (i->ioc_cmd == DLIOCRAW)
			printf(" = DLIOCRAW");
		else if (i->ioc_cmd == DL_IOC_HDR_INFO) {
			h = ioc_hdr_info_handler;
			printf(" = DL_IOC_HDR_INFO");
		}
		printf("\n");
		printf("    ioc_id 0x%x\n", i->ioc_id);
		printf("    ioc_flag 0x%x\n", i->ioc_flag);
		if (i->ioc_count == TRANSPARENT)
			printf("    ioc_count 0x%x == TRANSPARENT\n", 
				i->ioc_count);
		else
			printf("    ioc_count 0x%x\n", i->ioc_count);

		if (ty == M_IOCACK || ty == M_IOCNAK) {
			printf("    ioc_error %d\n", i->ioc_error);
			printf("    ioc_rval %d\n", i->ioc_rval);
		}
		return h;
	}

	if (ty != M_PROTO && ty != M_PCPROTO)
		return NULL;
	
	prim = *(unsigned *)p;
	switch (prim) {
	    case DL_INFO_REQ:
		printf("  DL_INFO_REQ\n");
		break;
	    case DL_INFO_ACK:
		{
		dl_info_ack_t *i = p;
		unsigned char *a;
		int n;

		printf("  DL_INFO_ACK\n");
		printf("    dl_max_sdu %lu\n", i->dl_max_sdu);
		printf("    dl_min_sdu %lu\n", i->dl_min_sdu);
		printf("    dl_addr_length %lu\n", i->dl_addr_length);
		printf("    dl_mac_type %lu = %s\n", 
			i->dl_mac_type, get_dl_mac_type(i->dl_mac_type));
		printf("    dl_current_state %lu\n", i->dl_current_state);
		printf("    dl_sap_length %ld\n", i->dl_sap_length);
		printf("    dl_service_mode %lu = %s\n", i->dl_service_mode, 
			get_dl_service_mode(i->dl_service_mode));
		printf("    dl_provider_style 0x%lx", i->dl_provider_style);
		switch (i->dl_provider_style) {
		    case DL_STYLE1: printf(" = DL_STYLE1\n"); break;
		    case DL_STYLE2: printf(" = DL_STYLE2\n"); break;
		    default: printf(" ???\n");
		}
		printf("    dl_addr_offset %lu", i->dl_addr_offset);
		if (i->dl_addr_offset > 0) {
			printf("\n");
			printf("    address");
			a = (unsigned char *)p + i->dl_addr_offset;
			for (n = 0; n < i->dl_addr_length; ++n, ++a)
				printf(" %02x", *a);
			printf("\n");
		}
		else
			printf(", unbound\n");
		printf("    dl_version %lu\n", i->dl_version);
		printf("    dl_brdcst_addr_length %lu\n", 
			i->dl_brdcst_addr_length);
		printf("    dl_brdcst_addr_offset %lu\n", 
			i->dl_brdcst_addr_offset);
		if (i->dl_brdcst_addr_offset > 0) {
			printf("    brdcast-address");
			a = (unsigned char *)p + i->dl_brdcst_addr_offset;
			for (n = 0; n < i->dl_brdcst_addr_length; ++n, ++a)
				printf(" %02x", *a);
			printf("\n");
		}
		break;
		}
	    case DL_ATTACH_REQ:
		{
		dl_attach_req_t *a = p;

		printf("  DL_ATTACH_REQ\n");
		printf("      dl_ppa %ld\n", a->dl_ppa);
		break;
		}
	    case DL_DETACH_REQ:
		printf("  DL_DETACH_REQ\n");
		break;
	    case DL_BIND_REQ:
		{
		dl_bind_req_t *b = p;

		printf("  DL_BIND_REQ\n");
		printf("    dl_sap 0x%lx = %s\n", 
			b->dl_sap, get_ether_type(b->dl_sap));
		printf("    dl_max_conind %lu\n", b->dl_max_conind);
		printf("    dl_service_mode %u = %s\n", b->dl_service_mode,
			get_dl_service_mode(b->dl_service_mode));
		printf("    dl_conn_mgmt %u\n", b->dl_conn_mgmt);
		printf("    dl_xidtest_flg 0x%lx\n", b->dl_xidtest_flg);
		break;
		}
	    case DL_BIND_ACK:
		{
		dl_bind_ack_t *b = p;
		unsigned char *a;
		int n;

		printf("  DL_BIND_ACK\n");
		printf("    dl_sap 0x%lx = %s\n", 
			b->dl_sap, get_ether_type(b->dl_sap));
		printf("    dl_addr_length %lu\n", b->dl_addr_length);
		printf("    dl_addr_offset %lu\n", b->dl_addr_offset);
		printf("    address");
		a = (unsigned char *)p + b->dl_addr_offset;
		for (n = 0; n < b->dl_addr_length; ++n, ++a)
			printf(" %02x", *a);
		printf("\n");
		printf("    dl_max_conind %lu\n", b->dl_max_conind);
		printf("    dl_xidtest_flg 0x%lx\n", b->dl_xidtest_flg);
		break;
		}
	    case DL_UNBIND_REQ:
		printf("  DL_UNBIND_REQ\n");
		break;
	    case DL_OK_ACK:
		{
		dl_ok_ack_t *o = p;

		printf("  DL_OK_ACK\n");
		printf("    dl_correct_primitive 0x%lx\n", 
			o->dl_correct_primitive);
		break;
		}
	    case DL_ERROR_ACK:
		{
		dl_error_ack_t *e = p;

		printf("  DL_ERROR_ACK\n");
		printf("    dl_error_primitive %lx\n", e->dl_error_primitive);
		printf("    dl_errno %lx\n", e->dl_errno);
		printf("    dl_unix_errno %lx\n", e->dl_unix_errno);
		break;
		}
	    case DL_SUBS_BIND_REQ:
		printf("  DL_SUBS_BIND_REQ\n");
		break;
	    case DL_SUBS_BIND_ACK:
		printf("  DL_SUBS_BIND_ACK\n");
		break;
	    case DL_SUBS_UNBIND_REQ:
		printf("  DL_SUBS_UNBIND_REQ\n");
		break;
	    case DL_ENABMULTI_REQ:
		{
		dl_enabmulti_req_t *m = p;
		unsigned char *a;
		int n;

		printf("  DL_ENABMULTI_REQ\n");
		printf("    dl_addr_length %lu\n", m->dl_addr_length);
		printf("    dl_addr_offset %lu\n", m->dl_addr_offset);
		printf("    address");
		a = (unsigned char *)p + m->dl_addr_offset;
		for (n = 0; n < m->dl_addr_length; ++n, ++a)
			printf(" %02x", *a);
		printf("\n");
		break;
		}
	    case DL_DISABMULTI_REQ:
		printf("  DL_DISABMULTI_REQ\n");
		break;
	    case DL_PROMISCON_REQ:
		{
		dl_promiscon_req_t *pr = p;

		printf("  DL_PROMISCON_REQ\n");
		printf("    dl_level %ld = %s\n", pr->dl_level, 
			get_dl_level( pr->dl_level));
		break;
		}
	    case DL_PROMISCOFF_REQ:
		{
		dl_promiscoff_req_t *pr = p;

		printf("  DL_PROMISCOFF_REQ\n");
		printf("    dl_level %ld = %s\n", pr->dl_level, 
			get_dl_level( pr->dl_level));
		break;
		}
	    case DL_UNITDATA_REQ:
		unitdata_req_dump(p);
		break;
	    case DL_UNITDATA_IND:
		{
		dl_unitdata_ind_t *u = p;
		int n;

		printf("  DL_UNITDATA_IND\n");
		printf("    dl_dest_addr_length %lu\n", u->dl_dest_addr_length);
		printf("    dl_dest_addr_offset %lu\n", u->dl_dest_addr_offset);
		nbytes_hex("    dest_address ", 
			(unsigned char *)p + u->dl_dest_addr_offset, 
				u->dl_dest_addr_length, "\n");
		printf("    dl_src_addr_length %lu\n", u->dl_src_addr_length);
		printf("    dl_src_addr_offset %lu\n", u->dl_src_addr_offset);
		nbytes_hex("    src_address ", 
			(unsigned char *)p + u->dl_src_addr_offset, 
				u->dl_src_addr_length, "\n");
		printf("    dl_group_address %lu\n", u->dl_group_address);

		break;
		}
	    case DL_UDERROR_IND:
		printf("  DL_UDERROR_IND\n");
		break;
	    case DL_UDQOS_REQ:
		printf("  DL_UDQOS_REQ\n");
		break;
	    case DL_CONNECT_REQ:
		printf("  DL_CONNECT_REQ\n");
		break;
	    case DL_CONNECT_IND:
		printf("  DL_CONNECT_IND\n");
		break;
	    case DL_CONNECT_RES:
		printf("  DL_CONNECT_RES\n");
		break;
	    case DL_CONNECT_CON:
		printf("  DL_CONNECT_CON\n");
		break;
	    case DL_TOKEN_REQ:
		printf("  DL_TOKEN_REQ\n");
		break;
	    case DL_TOKEN_ACK:
		printf("  DL_TOKEN_ACK\n");
		break;
	    case DL_DISCONNECT_REQ:
		printf("  DL_DISCONNECT_REQ\n");
		break;
	    case DL_DISCONNECT_IND:
		printf("  DL_DISCONNECT_IND\n");
		break;
	    case DL_RESET_REQ:
		printf("  DL_RESET_REQ\n");
		break;
	    case DL_RESET_IND:
		printf("  DL_RESET_IND\n");
		break;
	    case DL_RESET_RES:
		printf("  DL_RESET_RES\n");
		break;
	    case DL_RESET_CON:
		printf("  DL_RESET_CON\n");
		break;
	    case DL_DATA_ACK_REQ:
		printf("  DL_DATA_ACK_REQ\n");
		break;
	    case DL_DATA_ACK_IND:
		printf("  DL_DATA_ACK_IND\n");
		break;
	    case DL_DATA_ACK_STATUS_IND:
		printf("  DL_DATA_ACK_STATUS_IND\n");
		break;
	    case DL_REPLY_REQ:
		printf("  DL_REPLY_REQ\n");
		break;
	    case DL_REPLY_IND:
		printf("  DL_REPLY_IND\n");
		break;
	    case DL_REPLY_STATUS_IND:
		printf("  DL_REPLY_STATUS_IND\n");
		break;
	    case DL_REPLY_UPDATE_REQ:
		printf("  DL_REPLY_UPDATE_REQ\n");
		break;
	    case DL_REPLY_UPDATE_STATUS_IND:
		printf("  DL_REPLY_UPDATE_STATUS_IND\n");
		break;
	    case DL_XID_REQ:
		printf("  DL_XID_REQ\n");
		break;
	    case DL_XID_IND:
		printf("  DL_XID_IND\n");
		break;
	    case DL_XID_RES:
		printf("  DL_XID_RES\n");
		break;
	    case DL_XID_CON:
		printf("  DL_XID_CON\n");
		break;
	    case DL_TEST_REQ:
		printf("  DL_TEST_REQ\n");
		break;
	    case DL_TEST_IND:
		printf("  DL_TEST_IND\n");
		break;
	    case DL_TEST_RES:
		printf("  DL_TEST_RES\n");
		break;
	    case DL_TEST_CON:
		printf("  DL_TEST_CON\n");
		break;
	    case DL_PHYS_ADDR_REQ:
		{
		dl_phys_addr_req_t *pa = p;

		printf("  DL_PHYS_ADDR_REQ\n");
		printf("    dl_addr_type %lu", pa->dl_addr_type);
		switch (pa->dl_addr_type) {
		    case DL_FACT_PHYS_ADDR: 
			printf(" = DL_FACT_PHYS_ADDR\n");
			break;
		    case DL_CURR_PHYS_ADDR: 
			printf(" = DL_CURR_PHYS_ADDR\n");
			break;
		    default:
			printf(" = ???\n");
			break;
		}
		break;
		}
	    case DL_PHYS_ADDR_ACK:
		{
		dl_phys_addr_ack_t *pa = p;
		unsigned char *a;

		printf("  DL_PHYS_ADDR_ACK\n");
		printf("    dl_addr_length %lu\n", pa->dl_addr_length);
		printf("    dl_addr_offset %lu\n", pa->dl_addr_offset);
		a = (unsigned char *)p + pa->dl_addr_offset;
		nbytes_hex("    address ", a, pa->dl_addr_length, "\n");
		break;
		}
	    case DL_SET_PHYS_ADDR_REQ:
		printf("  DL_SET_PHYS_ADDR_REQ\n");
		break;
	    case DL_GET_STATISTICS_REQ:
		printf("  DL_GET_STATISTICS_REQ\n");
		break;
	    case DL_GET_STATISTICS_ACK:
		printf("  DL_GET_STATISTICS_ACK\n");
		break;
	    default:
		printf("unknown dlpi primitive %x\n", prim);
		break;
	}

	return NULL;
}

void 
dump_instance_sap(void *p, int ty, int l, t_uscalar_t dl_sap)
{
	if (ty != M_DATA)
		return;

	printf("    dl_sap 0x%x = %s\n", dl_sap, get_ether_type(dl_sap));
	switch (dl_sap) {
	    case 0: 		
		dump_ether_packet((struct ether_header *)p, l);
		break;
	    case ETHERTYPE_IP: 		
		dump_ip_packet((struct ip *)p, l);
		break;
	    case ETHERTYPE_ARP:		
		break;
	    case ETHERTYPE_REVARP:	
		break;
	}
}

void 
ioc_hdr_info_handler(void *p, int ty, int l, t_uscalar_t dl_sap)
{
	if (ty == M_PROTO)
		unitdata_req_dump(p);
	else if (ty == M_DATA) /* of M_IOCACK */
		ether_header_dump(p);
}

void
ether_header_dump(struct ether_header *p)
{
	int t;

	nbytes_hex("    ether_dhost ",p->ether_dhost.ether_addr_octet, 6, "\n");
	nbytes_hex("    ether_shost ",p->ether_shost.ether_addr_octet, 6, "\n");
	t = ntohs(p->ether_type);
	printf("    ether_type 0x%x = %s\n", t, get_ether_type(t));
}

void 
dump_ether_packet(struct ether_header *p, int l)
{
	int t;

	printf("  -- ETHER --\n");
	if (l < sizeof(struct ether_header))
		return;
	ether_header_dump(p);
	t = ntohs(p->ether_type);
	switch (t) {
	    case ETHERTYPE_IP:
		dump_ip_packet((struct ip *)(p + 1), 
			l - sizeof(struct ether_header));
		break;
	}
		
}

void
dump_ip_packet(struct ip *p, int l)
{
	printf("  -- IP --\n");
	if (l < sizeof(struct ip))
		return;
	dump_in_addr("    ip_src ", p->ip_src, "\n");
	dump_in_addr("    ip_dst ", p->ip_dst, "\n");
	printf("    ip_p %d = %s\n", p->ip_p, getipp( p->ip_p));
}

void
dump_in_addr(char *s1, struct in_addr a, char *s2)
{
	if (s1)
		printf(s1);
	printf("%d.%d.%d.%d", a._S_un._S_un_b.s_b1, a._S_un._S_un_b.s_b2,
		a._S_un._S_un_b.s_b3, a._S_un._S_un_b.s_b4);
	if (s2)
		printf(s2);
}

void 
nbytes_hex(char *s1, uchar_t *b, int n, char *s2)
{
	int k;

	if (s1)
		printf(s1);
	for (k = 0; k < n; ++k)
		printf("%02x ", b[k]);
	if (s2)
		printf(s2);
}

void
Error(char *str, ...)
{
        va_list ap;
	int e;

	e = errno;

        va_start(ap, str);
        vfprintf(stderr, str, ap);
	va_end(ap);
        fprintf( stderr, "\n" );

	if (e)
		fprintf(stderr, "system error %d: %s\n", e, strerror(e));

        exit(1);
}

void
unitdata_req_dump(dl_unitdata_req_t *p)
{
	int n;

	printf("  DL_UNITDATA_REQ\n");
	printf("    dl_dest_addr_length %lu\n", p->dl_dest_addr_length);
	printf("    dl_dest_addr_offset %lu\n", p->dl_dest_addr_offset);
	nbytes_hex("    address ", (unsigned char *)p + p->dl_dest_addr_offset, 
		p->dl_dest_addr_length, "\n");
	printf("    dl_priority %lu\n", p->dl_priority);
}

char *
getop(int t)
{
	switch (t) {
	    case SSPY_OP_DRV_OPEN:	return "DRV_OPEN";
	    case SSPY_OP_MOD_OPEN:	return "MOD_OPEN";
	    case SSPY_OP_CLOSE:		return "CLOSE";
	    case SSPY_OP_WPUT:		return "WPUT";
	    case SSPY_OP_RPUT:		return "RPUT";
	    default:			return "???";
	}
}

char *
gettype(int t)
{
	switch (t) {
	    case M_DATA:	return "M_DATA";
	    case M_PROTO:	return "M_PROTO";
	    case M_BREAK:	return "M_BREAK";
	    case M_PASSFP:	return "M_PASSFP";
	    case M_EVENT:	return "M_EVENT";
	    case M_SIG:		return "M_SIG";
	    case M_DELAY:	return "M_DELAY";
	    case M_CTL:		return "M_CTL";
	    case M_IOCTL:	return "M_IOCTL";
	    case M_SETOPTS:	return "M_SETOPTS";
	    case M_RSE:		return "M_RSE";
	    case M_IOCACK:	return "M_IOCACK";
	    case M_IOCNAK:	return "M_IOCNAK";
	    case M_PCPROTO:	return "M_PCPROTO";
	    case M_PCSIG:	return "M_PCSIG";
	    case M_READ:	return "M_READ";
	    case M_FLUSH:	return "M_FLUSH";
	    case M_STOP:	return "M_STOP";
	    case M_START:	return "M_START";
	    case M_HANGUP:	return "M_HANGUP";
	    case M_ERROR:	return "M_ERROR";
	    case M_COPYIN:	return "M_COPYIN";
	    case M_COPYOUT:	return "M_COPYOUT";
	    case M_IOCDATA:	return "M_IOCDATA";
	    case M_PCRSE:	return "M_PCRSE";
	    case M_STOPI:	return "M_STOPI";
	    case M_STARTI:	return "M_STARTI";
	    case M_PCEVENT:	return "M_PCEVENT";
	    case M_UNHANGUP:	return "M_UNHANGUP";
	    default: return "???";
	}
}

char *
getipp(int prot)
{
	switch (prot) {
	    case IPPROTO_HOPOPTS:	return "IPPROTO_HOPOPTS";
	    case IPPROTO_ICMP:		return "IPPROTO_ICMP";
	    case IPPROTO_IGMP:		return "IPPROTO_IGMP";
	    case IPPROTO_GGP:		return "IPPROTO_GGP";
	    case IPPROTO_ENCAP:		return "IPPROTO_ENCAP";
	    case IPPROTO_TCP:		return "IPPROTO_TCP";
	    case IPPROTO_EGP:		return "IPPROTO_EGP";
	    case IPPROTO_PUP:		return "IPPROTO_PUP";
	    case IPPROTO_UDP:		return "IPPROTO_UDP";
	    case IPPROTO_IDP:		return "IPPROTO_IDP";
	    case IPPROTO_IPV6:		return "IPPROTO_IPV6";
	    case IPPROTO_ROUTING:	return "IPPROTO_ROUTING";
	    case IPPROTO_FRAGMENT:	return "IPPROTO_FRAGMENT";
	    case IPPROTO_RSVP:		return "IPPROTO_RSVP";
	    case IPPROTO_ESP:		return "IPPROTO_ESP";
	    case IPPROTO_AH:		return "IPPROTO_AH";
	    case IPPROTO_ICMPV6:	return "IPPROTO_ICMPV6";
	    case IPPROTO_NONE:		return "IPPROTO_NONE";
	    case IPPROTO_DSTOPTS:	return "IPPROTO_DSTOPTS";
	    case IPPROTO_HELLO:		return "IPPROTO_HELLO";
	    case IPPROTO_ND:		return "IPPROTO_ND";
	    case IPPROTO_EON:		return "IPPROTO_EON";
	    case IPPROTO_PIM:		return "IPPROTO_PIM";
	    case IPPROTO_RAW:		return "IPPROTO_RAW";
	    default: return "???";
	}
}

char *
get_dl_mac_type(t_uscalar_t ty)
{
	switch (ty) {
	    case DL_CSMACD:	return "DL_CSMACD";
	    case DL_TPB:	return "DL_TPB";
	    case DL_TPR:	return "DL_TPR";
	    case DL_METRO: 	return "DL_METRO";
	    case DL_ETHER: 	return "DL_ETHER";
	    case DL_HDLC: 	return "DL_HDLC";
	    case DL_CHAR: 	return "DL_CHAR";
	    case DL_CTCA: 	return "DL_CTCA";
	    case DL_FDDI: 	return "DL_FDDI";
	    case DL_FC: 	return "DL_FC";
	    case DL_ATM: 	return "DL_ATM";
	    case DL_IPATM: 	return "DL_IPATM";
	    case DL_X25: 	return "DL_X25";
	    case DL_ISDN: 	return "DL_ISDN";
	    case DL_HIPPI: 	return "DL_HIPPI";
	    case DL_100VG: 	return "DL_100VG";
	    case DL_100VGTPR: 	return "DL_100VGTPR";
	    case DL_ETH_CSMA: 	return "DL_ETH_CSMA";
	    case DL_100BT: 	return "DL_100BT";
	    case DL_FRAME: 	return "DL_FRAME";
	    case DL_MPFRAME: 	return "DL_MPFRAME";
	    case DL_ASYNC: 	return "DL_ASYNC";
	    case DL_IPX25: 	return "DL_IPX25";
	    case DL_LOOP: 	return "DL_LOOP";
	    case DL_OTHER: 	return "DL_OTHER";
	    default: return "???";
	}
}

char *
get_dl_service_mode(t_uscalar_t m)
{
	switch (m) {
	    case DL_CODLS:	return "DL_CODLS";
	    case DL_CLDLS:	return "DL_CLDLS";
	    case DL_ACLDLS:	return "DL_ACLDLS";
	    default: return "???";
	}
}

char *
get_ether_type(int ty)
{
	switch (ty) {
	    case 0:			return "NULL";
	    case ETHERTYPE_PUP:		return "ETHERTYPE_PUP";
	    case ETHERTYPE_IP:		return "ETHERTYPE_IP";
	    case ETHERTYPE_ARP:		return "ETHERTYPE_ARP";
	    case ETHERTYPE_REVARP:	return "ETHERTYPE_REVARP";
	    case ETHERTYPE_IPV6:	return "ETHERTYPE_IPV6";
	    default: 			return "???";
	}
}

char *
get_dl_level(int l)
{
	switch (l) {
	    case DL_PROMISC_PHYS:	return "DL_PROMISC_PHYS";
	    case DL_PROMISC_SAP:	return "DL_PROMISC_SAP";
	    case DL_PROMISC_MULTI:	return "DL_PROMISC_MULTI";
	    default: 			return "???";
	}
}
