#include <sys/types.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/dlpi.h>

#include "sspy.h"
#include <sys/ddi.h>
#include <sys/sunddi.h>

static int sspy_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd,  void
     *arg, void **resultp);
static int sspy_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int sspy_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int sspy_open(queue_t *q, dev_t *devp, int oflag, 
					int sflag, cred_t *credp);
static int sspy_close(queue_t *q, int oflag, cred_t *credp);

static int sspy_uwput(queue_t *q, mblk_t *mp);
static int sspy_uwsrv(queue_t *q);
static int sspy_lwsrv(queue_t *q);

static int sspy_urput(queue_t *q, mblk_t *mp);
static int sspy_ursrv(queue_t *q);
static int sspy_lrput(queue_t *q, mblk_t *mp);
static int sspy_lrsrv(queue_t *q);

static struct module_info minfo = {0x123, "sspy", 0, INFPSZ, STRLOW, STRHIGH};

static struct qinit uwinit = {
	sspy_uwput, sspy_uwsrv, NULL, NULL, NULL, &minfo, NULL,
};

static struct qinit urinit = {
	sspy_urput, sspy_ursrv, sspy_open, sspy_close, NULL, &minfo, NULL,
};

static struct qinit lwinit = {
	NULL, sspy_lwsrv, NULL, NULL, NULL, &minfo, NULL,
};

static struct qinit lrinit = {
	sspy_lrput, sspy_lrsrv, NULL, NULL, NULL, &minfo, NULL,
};

struct streamtab sspy_strtab = {
	&urinit, &uwinit, &lrinit, &lwinit, 
};

static struct cb_ops sspy_cb_ops = {
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nodev,
	nochpoll,		/* cb_chpoll 		*/
	ddi_prop_op,		/* cb_prop_op 		*/
	&sspy_strtab,		/* cb_str, streamtab 	*/
	D_NEW|D_MP|D_MTPERMOD,	/* cb_flag 		*/
	CB_REV,			/* cb_rev 		*/
	nodev,
	nodev
};

static struct dev_ops sspy_ops = {
	DEVO_REV,		/* devo_rev		*/
	0,			/* ref count 		*/
	sspy_getinfo,		/* devo_getinfo		*/
	nulldev,		/* devo_identify	*/
	nulldev,		/* devo_probe		*/
	sspy_attach,		/* devo_attach		*/
	sspy_detach,		/* devo_detach		*/
	nodev,			/* devo_reset		*/
	&sspy_cb_ops,		/* devo_cb_ops		*/
	NULL,			/* devo_bus_ops		*/
	nulldev			/* devo_power		*/
};

static struct modldrv md_drv = {
	&mod_driverops, 	/* Type of module. This is a driver */
	"sspy driver v0.1", 	/* Name of the module */
	&sspy_ops,
};

static struct fmodsw sspy_fmodsw = {
	"sspy",
	&sspy_strtab,
	D_NEW|D_MP|D_MTPERMOD
};

static struct modlstrmod md_mod = {
	&mod_strmodops, "sspy module v0.1", &sspy_fmodsw
};

static struct modlinkage ml = {
	MODREV_1,
	&md_drv,	/* the driver	*/
	&md_mod,	/* the module	*/
	NULL
};

/* device to instance mapping: 
 *	There is one static instance. Minor 1 is clonable and will send
 *	trace records of all i/o to minor 0 which is not clonable
 *	and is the control device.
 */
#define  SSPY_DEV_TO_INST(dev)			(getminor(dev) >> 1)
#define  SSPY_DEV_TO_TYPE(dev)			(getminor(dev) & 1)
#define  SSPY_INST_TO_DEV(instance, type)	((instance) << 1 | (type))
#define  DEV_TYPE_CTL	0
#define  DEV_TYPE_CLONE	1

static int sspy_pflag = 1;

#define SSPY_QBOT_MAX	5

struct qbot {
	int	in_use;
	queue_t	*q;			/* write q			*/
};

struct sspy_ctl {
	dev_info_t	*dip;		/* device info pointer		*/
	queue_t 	*q;		/* pointer to read queue 	*/
	struct qbot	qbot[SSPY_QBOT_MAX]; /* all I_[P]LINK'ed streams*/
};

static struct sspy_ctl sspy_ctl;	/* the control device 		*/

struct sspy_clone {
	int		in_use;
	int		close_flush;	/* flush on close pending	*/
	queue_t 	*q;		/* pointer to read queue 	*/
	queue_t 	*l_qbot;	/* pointer to lower write queue	*/
	int		qbot_idx;	/* index into sspy_ctl.qbot[]	*/
	struct dlpi_info dlpi_info;	/* dlpi bind info		*/
};

#define SSPY_CLONE_MAX 10
static struct sspy_clone sspy_clones[SSPY_CLONE_MAX]; /* all clones	*/

static unsigned sspy_trace_seqno = 0;	/* sequence number for trace	*/


static int sspy_ctl_open(queue_t *q, dev_t *devp, int oflag, 
					int sflag, cred_t *credp);
static int sspy_clone_open(queue_t *q, dev_t *devp, int oflag, 
					int sflag, cred_t *credp);
static int sspy_module_open(queue_t *q, dev_t *devp, int oflag, 
					int sflag, cred_t *credp);
static int sspy_ctl_wput(queue_t *q, mblk_t *mp);
static void sspy_do_ioctl(queue_t *q, mblk_t *mp);
static void sspy_trace_out(enum sspy_op op, queue_t *q, mblk_t *mp, int err);
static int sspy_packmsg(mblk_t *from, mblk_t *to);


int
_init(void)
{
	int e;

	if (sspy_pflag) 
		cmn_err(CE_CONT, "sspy: _init()");

	e = mod_install(&ml);
	if (e)
		return e;

	return 0;
}

int
_info(struct modinfo *modinfop)
{
	if (sspy_pflag) 
		cmn_err(CE_CONT, "sspy: _info()");

	return mod_info(&ml, modinfop);
}

int
_fini(void)
{
	int e;

	if (sspy_pflag) 
		cmn_err(CE_CONT, "sspy: _fini()");
	
	return(mod_remove(&ml));
}

static int
sspy_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd,  void *arg, void **resultp)
{
	int instance;

	*resultp = NULL;
	instance = SSPY_DEV_TO_INST((dev_t)arg);

	switch (cmd) {
	    case DDI_INFO_DEVT2DEVINFO:
		if (sspy_pflag) 
			cmn_err(CE_CONT, 
				"sspy%d: sspy_getinfo(DDI_INFO_DEVT2DEVINFO)",
			    	instance);

		*resultp = sspy_ctl.dip;
		return DDI_SUCCESS;

	    case DDI_INFO_DEVT2INSTANCE: /* gerufen bei deferred attach */
		if (sspy_pflag) 
			cmn_err(CE_CONT, 
			"sspy%d: sspy_getinfo(DDI_INFO_DEVT2INSTANCE)",
			    	instance);
		*resultp = (void *)instance;
		return DDI_SUCCESS;

	    default:
		return DDI_FAILURE;
	}
}

static int
sspy_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int	instance;
	minor_t	minor;

	instance = ddi_get_instance(dip);

	if (sspy_pflag) 
		cmn_err(CE_CONT, "sspy%d: attach()", instance);

	switch (cmd) {
	    case DDI_ATTACH:
		if (sspy_pflag) 
			cmn_err(CE_CONT, 
				"sspy%d: sspy_attach(DDI_ATTACH)", instance);

		minor = SSPY_INST_TO_DEV(instance, DEV_TYPE_CTL);
		if (ddi_create_minor_node(dip, "sspy", S_IFCHR, minor,
			DDI_PSEUDO, 0) != DDI_SUCCESS) {
				cmn_err(CE_NOTE,
				    "sspy%d attach: could not create minor %d",
				    instance, minor);
				return DDI_FAILURE;
		}
		minor = SSPY_INST_TO_DEV(instance, DEV_TYPE_CLONE);
		if (ddi_create_minor_node(dip, "sspymux", S_IFCHR, minor,
			DDI_PSEUDO, CLONE_DEV) != DDI_SUCCESS) {
				cmn_err(CE_NOTE,
				    "sspy%d attach: could not create minor %d",
				    instance, minor);
				ddi_remove_minor_node(dip, NULL);
				return DDI_FAILURE;
		}

		bzero(&sspy_ctl, sizeof(struct sspy_ctl));
		sspy_ctl.dip = dip;

		ddi_report_dev(dip);
		return (DDI_SUCCESS);
	    default:
		if (sspy_pflag) 
			cmn_err(CE_CONT, 
				"sspy%d: sspy_attach(%d)", instance, cmd);
		return DDI_FAILURE;
	}
}

static int
sspy_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int instance, qbot_idx;

	instance = ddi_get_instance(dip);

	if (sspy_pflag) 
		cmn_err(CE_CONT, "sspy%d: dettach()", instance);

	switch (cmd) {
	    case DDI_DETACH:
		if (sspy_pflag) 
			cmn_err(CE_CONT, 
				"sspy: sspy_detach(DDI_DETACH), i%d", instance);

		/* Without this permanently linked lower streams stay open
		 * without a reference to unlink them.
		 */
		for (qbot_idx = 0; qbot_idx < SSPY_QBOT_MAX; ++qbot_idx) {
			if (sspy_ctl.qbot[qbot_idx].q) {
				cmn_err(CE_WARN, "sspy: cannot detach");
				cmn_err(CE_CONT, 
					"first unlink all lower streams");
				return DDI_FAILURE;
			}
		}

		ddi_remove_minor_node(dip, NULL);

		return (DDI_SUCCESS);

	    default:
		if (sspy_pflag) 
			cmn_err(CE_CONT, 
				"sspy%d: sspy_detach(%d)", instance, cmd);
		return DDI_FAILURE;
	}
}

static int 
sspy_open(queue_t *q, dev_t *devp, int oflag, int sflag, cred_t *credp)
{
	int instance, e;

	if (sspy_pflag) 
		cmn_err(CE_CONT, "sspy: open()");
	
	instance = SSPY_DEV_TO_INST(*devp);
	if (sflag != MODOPEN && instance != 0) {
		cmn_err(CE_NOTE, "sspy%d: open()", instance);
		return ENOENT;
	}

	if (!sflag) {
		/* open of control device */

		if (SSPY_DEV_TO_TYPE(*devp) != DEV_TYPE_CTL) {
		    cmn_err(CE_NOTE,"sspy%d: open(), bad minor %d", 
			instance, getminor(*devp));
		    return ENOENT;
		}
		e = sspy_ctl_open(q, devp, oflag, sflag, credp);
		if (e)
			return e;
	}
	else if (sflag == CLONEOPEN) {
		/* open of clone device */

		if (q->q_ptr) {
			cmn_err(CE_NOTE, "sspy%d: bad clone open()", instance);
			return ENOENT;
		}
		e = sspy_clone_open(q, devp, oflag, sflag, credp);
		if (e)
			return e;
	}
	else if (sflag == MODOPEN) {
		/* open of module */

		if (q->q_ptr) {
			cmn_err(CE_NOTE, "sspy%d: bad module open()", instance);
			return ENOENT;
		}
		e = sspy_module_open(q, devp, oflag, sflag, credp);
		if (e)
			return e;
	}
	else
		return EINVAL;

	if (sspy_pflag) {
		instance = SSPY_DEV_TO_INST(*devp);
		cmn_err(CE_CONT, "sspy opened instance %d, minor %d%s",
			instance, getminor(*devp),
			sflag == CLONEOPEN ? " CLONEOPEN" : "");
	}

	return 0;
}

static int 
sspy_ctl_open(queue_t *q, dev_t *devp, int oflag, int sflag, cred_t *credp)
{
	if (q->q_ptr)
		return 0;	/* already open */

	sspy_ctl.q = q;
	WR(q)->q_ptr = q->q_ptr = (queue_t *)&sspy_ctl;
	
	qprocson(q);

	return 0;
}

static int 
sspy_clone_open(queue_t *q, dev_t *devp, int oflag, int sflag, cred_t *credp)
{
	int instance, qbot_idx;
	int err = 0;
	queue_t *h;

	for (instance = 0; instance < SSPY_CLONE_MAX; ++ instance)
		if (!sspy_clones[instance].in_use)
			break;
	if (instance == SSPY_CLONE_MAX) {
		err = EMFILE;
		goto out;
	}
	
	for (qbot_idx = 0; qbot_idx < SSPY_QBOT_MAX; ++ qbot_idx)
		if (!sspy_ctl.qbot[qbot_idx].in_use && 
				sspy_ctl.qbot[qbot_idx].q)
			break;
	if (qbot_idx == SSPY_QBOT_MAX) {
		err = EXDEV;
		goto out;
	}

	*devp = makedevice(getmajor(*devp), 
		SSPY_INST_TO_DEV(instance, DEV_TYPE_CLONE));

	bzero(sspy_clones + instance, sizeof(struct sspy_clone));
	sspy_clones[instance].in_use = 1;
	sspy_clones[instance].q = q;
	sspy_ctl.qbot[qbot_idx].in_use = 1;
	h = sspy_clones[instance].l_qbot = sspy_ctl.qbot[qbot_idx].q;
	sspy_clones[instance].qbot_idx = qbot_idx;
	sspy_clones[instance].dlpi_info.addr_len = SSPY_DLPI_UNBOUND;
	WR(q)->q_ptr = q->q_ptr = sspy_clones + instance;

	RD(h)->q_ptr = h->q_ptr = sspy_clones + instance;

	qprocson(q);

out:

	sspy_trace_out(SSPY_OP_DRV_OPEN, q, NULL, err);

	return err;
}

static int 
sspy_module_open(queue_t *q, dev_t *devp, int oflag, int sflag, cred_t *credp)
{
	int instance;
	int err = 0;

	for (instance = 0; instance < SSPY_CLONE_MAX; ++ instance)
		if (!sspy_clones[instance].in_use)
			break;
	if (instance == SSPY_CLONE_MAX) {
		err = EMFILE;
		goto out;
	}
	
	bzero(sspy_clones + instance, sizeof(struct sspy_clone));
	sspy_clones[instance].in_use = 1;
	sspy_clones[instance].q = q;
	sspy_clones[instance].l_qbot = NULL; /* this is a module */
	sspy_clones[instance].qbot_idx = -1;
	WR(q)->q_ptr = q->q_ptr = sspy_clones + instance;

	qprocson(q);

out:

	sspy_trace_out(SSPY_OP_MOD_OPEN, q, NULL, err);

	return err;
}

static int 
sspy_close(queue_t *q, int oflag, cred_t *credp)
{
	struct sspy_clone *c;
	mblk_t *mp;

	if (!q->q_ptr)
		return ENODEV;

	/* close of control device */
	if (q->q_ptr == &sspy_ctl) {
		if (sspy_pflag) 
			cmn_err(CE_CONT, "sspy: closing control device");

		qprocsoff(q);
		sspy_ctl.q = NULL;
		q->q_ptr = WR(q)->q_ptr = NULL;
		return 0;
	}

	/* close of clone device or module */
	sspy_trace_out(SSPY_OP_CLOSE, q, NULL, 0);

	c = q->q_ptr;

	if (sspy_pflag) 
		cmn_err(CE_CONT, "sspy: closing clone %d", c - sspy_clones);

	/* For clone driver close set close_flush and flush lower queues.
	 * See sspy_lrput.
	 */
	if (c->l_qbot) {
		c->close_flush = 1;
		putnextctl1(c->l_qbot, M_FLUSH, FLUSHRW);
		do
			qwait(q);
		while (c->close_flush);
	}

	qprocsoff(q);

	if (c->l_qbot) {
		sspy_ctl.qbot[c->qbot_idx].in_use = 0;
		c->l_qbot->q_ptr = NULL;
		c->l_qbot = NULL;
	}
	c->in_use = 0;
	c->q = NULL;

	return 0;
}

static int 
sspy_uwput(queue_t *q, mblk_t *mp)
{
	struct sspy_clone *c;

	/* control device */
	if (q->q_ptr == &sspy_ctl)
		return sspy_ctl_wput(q, mp);

	/* clone driver or module */
	c = q->q_ptr;

	if (mp->b_datap->db_type == M_FLUSH) {
		/* canonical flush handling, read q is unused */
		if (*mp->b_rptr & FLUSHW) {
			/* flush write q */
			if (*mp->b_rptr & FLUSHBAND)
				flushband(q, FLUSHDATA, *(mp->b_rptr+1));
			else
				flushq(q, FLUSHDATA);
		}
	}

	sspy_trace_out(SSPY_OP_WPUT, q, mp, 0);

	if (!c->l_qbot) {  /* module */
		putnext(q, mp);
		return 0;
	}

	if (mp->b_datap->db_type >= QPCTL)
		putnext(c->l_qbot, mp);
	else if (q->q_first || !canputnext(c->l_qbot))
		putq(q, mp);
	else
		putnext(c->l_qbot, mp);

	return 0;
}

/* 
 * Write q put for control device. No messages are put into the q.
 */
static int 
sspy_ctl_wput(queue_t *q, mblk_t *mp)
{
	switch (mp->b_datap->db_type) {
	    case M_IOCTL:
		sspy_do_ioctl(q, mp);
		break;

	    case M_FLUSH:
		/* canonical flush handling */
		if (*mp->b_rptr & FLUSHW) {
			/* flush write q */
			if (*mp->b_rptr & FLUSHBAND)
				flushband(q, FLUSHDATA, *(mp->b_rptr+1));
			else
				flushq(q, FLUSHDATA);
		}
		if (*mp->b_rptr & FLUSHR) {
			/* send request back upstream without FLUSHW */
			*mp->b_rptr &= ~FLUSHW;
			qreply(q, mp);
		}
		else
			freemsg(mp); /* free message */
		break;
	
	    default:
		freemsg(mp);
		break;
	}

	return 0;
}

static void 
sspy_do_ioctl(queue_t *q, mblk_t *mp)
{
	struct iocblk *ioc;
	struct linkblk *lblk;
	mblk_t *mp2;
	int n, err = EINVAL;

	ioc = (struct iocblk *) mp->b_rptr;
	if (ioc->ioc_count == TRANSPARENT) {
		err = EINVAL;
		goto ioc_nak;
	}
	switch (ioc->ioc_cmd) {
	    case I_LINK:
	    case I_PLINK:
	    case I_UNLINK:
	    case I_PUNLINK:
		if (ioc->ioc_count != sizeof(struct linkblk) ||
				!(mp2 = mp->b_cont) || 
		  		mp2->b_wptr - mp2->b_rptr != 
					sizeof(struct linkblk)) {
			err = EINVAL;
			goto ioc_nak;
		}
		lblk = (struct linkblk *)mp2->b_rptr;

		if (ioc->ioc_cmd == I_LINK || ioc->ioc_cmd == I_PLINK) {
			for (n = 0; n < SSPY_QBOT_MAX; ++n)
				if (!sspy_ctl.qbot[n].q)
					break;
			if (n == SSPY_QBOT_MAX) {
				err = EMFILE;
				goto ioc_nak;
			}
			sspy_ctl.qbot[n].in_use = 0;
			sspy_ctl.qbot[n].q = lblk->l_qbot;
			lblk->l_qbot->q_ptr = NULL;
			/* ioctl(fd, I_UNLINK, 0) -> EINVAL !! */
			lblk->l_index = n + 1;

			if (sspy_pflag) 
				cmn_err(CE_CONT, 
					"sspy ioctl: linked lower stream %d",
						lblk->l_index);
		}
		else { /* I_UNLINK, I_PUNLINK */
			n = lblk->l_index - 1;
			if (n < 0 || n >= SSPY_QBOT_MAX) {
				err = EINVAL;
				goto ioc_nak;
			}
			if (!sspy_ctl.qbot[n].q) {
				err = ENXIO;
				goto ioc_nak;
			}
			if (sspy_ctl.qbot[n].in_use) {
				err = EBUSY;
				goto ioc_nak;
			}

			sspy_ctl.qbot[n].q = NULL;

			if (sspy_pflag) 
				cmn_err(CE_CONT, 
					"sspy ioctl: unlinked lower stream %d",
						lblk->l_index);
		}
		break;

	    default:
		err = EINVAL;
		goto ioc_nak;
	}

	/* send ACK */
	mp->b_datap->db_type = M_IOCACK;
	ioc->ioc_error = ioc->ioc_rval = 0;
	qreply(q, mp);

	return;

ioc_nak:
	mp->b_datap->db_type = M_IOCNAK;
	ioc->ioc_error = err;
	qreply(q, mp);
}

/* 
 * Does not expect control device messages in ctl dev's queue.
 * Assumes it works on the clone device q.
 */
static int 
sspy_uwsrv(queue_t *q)
{
	mblk_t *mp;
	struct sspy_clone *c = q->q_ptr;

	while (mp = getq(q)) {
		if (!c->l_qbot) {
			freemsg(mp); /* should not happen */
			continue;
		}

		if (canputnext(c->l_qbot))
			putnext(c->l_qbot, mp);
		else {
			putbq(q, mp);
			return 0;
		}
	}

	return 0;
}

static int 
sspy_lwsrv(queue_t *q)
{
	struct sspy_clone *c;

	if (sspy_pflag) 
		cmn_err(CE_CONT, "sspy_lwsrv: backenabling upper write q");

	/* Called when q is backenabled. Backenable upper write q. */
	c = q->q_ptr;
	if (!c || !c->q)
		return 0; /* happens after last flush on close */
	qenable(WR(c->q));

	return 0;
}

/*
 * Only used by module.
 */
static int 
sspy_urput(queue_t *q, mblk_t *mp)
{
	struct sspy_clone *c;

	if (q->q_ptr == &sspy_ctl)
		cmn_err(CE_PANIC, "sspy_urput called by ctl device");

	c = q->q_ptr;
	if (c->l_qbot)
		cmn_err(CE_PANIC, "sspy_urput called by clone driver");

	if (mp->b_datap->db_type == M_FLUSH) {
		/* canonical flush handling, read q is unused */
		if (*mp->b_rptr & FLUSHR) {
			/* flush write q */
			if (*mp->b_rptr & FLUSHBAND)
				flushband(q, FLUSHDATA, *(mp->b_rptr+1));
			else
				flushq(q, FLUSHDATA);
		}
	}

	sspy_trace_out(SSPY_OP_RPUT, q, mp, 0);

	putnext(q, mp);

	return 0;
}

static int 
sspy_ursrv(queue_t *q)
{
	struct sspy_clone *c;

	if (sspy_pflag) 
		cmn_err(CE_CONT, "sspy_ursrv: backenabling lower read q");

	/* Called when q is backenabled. Backenable lower read q. */
	c = q->q_ptr;
	if (!c || !c->l_qbot)
		return 0;
	qenable(RD(c->l_qbot));

	return 0;
}


static int 
sspy_lrput(queue_t *q, mblk_t *mp)
{
	struct sspy_clone *c;

	c = q->q_ptr;

	if (mp->b_datap->db_type == M_FLUSH) {
		/* canonical flush handling, read q is unused */
		if (*mp->b_rptr & FLUSHR) {
			/* flush read q */
			if (*mp->b_rptr & FLUSHBAND)
				flushband(q, FLUSHDATA, *(mp->b_rptr+1));
			else
				flushq(q, FLUSHDATA);
		}
	}

	/* sspy_close sends a flush after setting close_flush. */
	if (c->close_flush) {
		freemsg(mp);
		c->close_flush = 0;
		return 0;
	}

	/* for dlpi drivers keep track of bind status by catching
	 * DL_BIND_ACK and unbind acks (DL_OK_ACK for DL_UNBIND_REQ).
	 */
	if (mp->b_datap->db_type == M_PCPROTO) {
		dl_bind_ack_t *ba = (dl_bind_ack_t *)mp->b_rptr;
		dl_ok_ack_t *oa = (dl_ok_ack_t *)mp->b_rptr;
		int sz = mp->b_wptr - mp->b_rptr;

		if (sz >= sizeof(dl_bind_ack_t) && 
				ba->dl_primitive == DL_BIND_ACK) {
			c->dlpi_info.dl_sap = ba->dl_sap;
			c->dlpi_info.addr_len = ba->dl_addr_length;
			bcopy((unsigned char *)ba + ba->dl_addr_offset,
				c->dlpi_info.addr, ba->dl_addr_length);
		}
		else if(sz >= sizeof(dl_ok_ack_t) && 
			oa->dl_primitive == DL_OK_ACK &&
			oa->dl_correct_primitive == DL_UNBIND_REQ) {
			c->dlpi_info.addr_len = SSPY_DLPI_UNBOUND;
		}
	}

	/* for dlpi drivers keep track of DLIOCRAW ioctl acks */
	if (mp->b_datap->db_type == M_IOCACK && 
	    mp->b_wptr - mp->b_rptr >= sizeof(struct iocblk)) {
		struct iocblk *ioc = (struct iocblk *) mp->b_rptr;
		if (ioc->ioc_cmd == DLIOCRAW)
			c->dlpi_info.flags |= SSPY_DLPI_DLIOCRAW;
	}

	sspy_trace_out(SSPY_OP_RPUT, q, mp, 0);

	if (mp->b_datap->db_type >= QPCTL)
		putnext(c->q, mp);
	else if(q->q_first || !canputnext(c->q))
		putq(q, mp);
	else
		putnext(c->q, mp);

	return 0;
}

static int 
sspy_lrsrv(queue_t *q)
{
	mblk_t *mp;
	struct sspy_clone *c;

	c = q->q_ptr;

	while (mp = getq(q)) {
		if (canputnext(c->q))
			putnext(c->q, mp);
		else {
			putbq(q, mp);
			return 0;
		}
	}

	return 0;
}

static void 
sspy_trace_out(enum sspy_op op, queue_t *q, mblk_t *mp, int err)
{
	struct sspy_clone *c = q->q_ptr;
	mblk_t *trc_mp, *mp2;
	struct sspy_trace_ctl *t;
	int len;

	++sspy_trace_seqno;

	if (!sspy_ctl.q || !canputnext(sspy_ctl.q))
		return;	/* control device not open or q full */

	if (!c)
		return; /* should not happen */
	
	if (!(trc_mp = allocb(sizeof(struct sspy_trace_ctl), BPRI_MED)))
		return;
	
	t = (struct sspy_trace_ctl *)trc_mp->b_rptr;
	trc_mp->b_wptr += sizeof(struct sspy_trace_ctl);
	trc_mp->b_datap->db_type = M_PROTO;
	t->op = op;
	t->inst = c - sspy_clones;
	t->err = err;
	t->seqno = sspy_trace_seqno;
	bcopy(&c->dlpi_info, &t->dlpi_info, sizeof(struct dlpi_info));

	if (mp) {
		len = sspy_packmsg(mp, NULL);
		if (len <= 0 || !(mp2 = allocb(len, BPRI_MED))) {
			freemsg(trc_mp);
			return;
		}
		if (sspy_packmsg(mp, mp2) <= 0) {
			freemsg(trc_mp);
			freemsg(mp2);
			return;
		}
		trc_mp->b_cont = mp2;
	}

	putnext(sspy_ctl.q, trc_mp);
}

/*
 * Pack message "from" into a single M_DATA message into "to". Return the
 * length of the packed message. If "to" is NULL no packing is none. Only
 * the required length is calculated and returned.
 */
static int
sspy_packmsg(mblk_t *from, mblk_t *to)
{
	mblk_t *mp;
	unsigned char *p_to_start, *p_to, *p, *p_to_end;
	int l;

	if (to) {
		p_to_start = p_to = to->b_rptr;
		p_to_end = to->b_datap->db_lim;
	}
	else
		p_to_start = p_to = p_to_end = NULL;

	for (mp = from; mp; mp = mp->b_cont) {
		l = mp->b_wptr - mp->b_rptr;
		p = SSP_MSG_LEN_NEXT(p_to, l);
		if (to) {
			if (p > p_to_end)
				return -1;
			SSP_MSG_TYPE(p_to) = mp->b_datap->db_type;
			SSP_MSG_LEN(p_to) = l;
			bcopy(mp->b_rptr, SSP_MSG_START(p_to), l);
		}
		p_to = p;
	}

	if (to)
		to->b_wptr = p_to;

	return p_to - p_to_start;
}
