/*
 * Uses /dev/sspy to trace streams messages.
 *
 * usage: xtrace [-a | -h]
 *
 *	  -a: Dump all message contents.
 *	  -h: Help.
 */

#include <sys/types.h>
#include <stropts.h>
#include <sys/conf.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/stream.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "sspy.h"

int aflag = 0;	/* -a option */
#define MAXDATA (64*1024)

void Error(char *str, ...);
char *gettype(int t);
char *getop(int t);
void xdump(unsigned char *p, int l);
void ascii_out(unsigned char *p, int n);

int
main(int argc, char **argv)
{
	int fd, flag, e, l;
	struct strbuf ctlbuf, databuf;
	struct sspy_trace_ctl ctl;
	unsigned char unaligned_data[MAXDATA+7], *p;
	unsigned char *data;

	setbuf(stdout, NULL);
	setbuf(stderr, NULL);

	if (argc > 1) {
		if (!strcmp(argv[1], "-h"))
			Error("usage: %s [-a | -h]", argv[0]);
		aflag = 1;
	}

	fd = open(SSPYDEV, O_RDWR);
	if (fd < 0)
		Error("cannot open %s", SSPYDEV);
	
	data = SSP_ALIGN8(unaligned_data);
	for (;;) {
		ctlbuf.buf = (char *)&ctl;
		ctlbuf.maxlen = sizeof(struct sspy_trace_ctl);
		databuf.buf = (char *)data;
		databuf.maxlen = MAXDATA;
		flag = 0;

		e = getmsg(fd, &ctlbuf, &databuf, &flag);
		if (e & (MORECTL|MOREDATA))
			Error("getmsg MORECTL|MOREDATA= 0x%x", e);
		if (e)
			Error("getmsg");
		
		printf("%s, i %d, seq %u, e %d, l %d\n", getop(ctl.op),
			ctl.inst, ctl.seqno, ctl.err, databuf.len);

		if (databuf.len < 0)
			continue;

		for (p = data; p < data + databuf.len; p = SSP_MSG_NEXT(p)) {
			l = SSP_MSG_LEN(p);
			printf("  %s, len %d\n", gettype(SSP_MSG_TYPE(p)), l);
			if (aflag)
				xdump((unsigned char *)SSP_MSG_START(p), l);
		}
	}
		
}

void
xdump(unsigned char *p, int l)
{
	int n;

	for (n = 0; n < l; ++n) {
		if (!(n & 0xf))
			printf("    %04x: ", n);
		if (!(n & 0x7))
			putchar(' ');
		printf("%02x ", p[n]);
		if ((n & 0xf) == 0xf)
			ascii_out(p, n);
	}

	if (n & 0xf)
		ascii_out(p, n-1);
}

void
ascii_out(unsigned char *p, int n)
{
	int i;

	for (i = (n & 0xf) + 1; i < 16; ++i) {
		if (i == 8)
			putchar(' ');
		printf("   ");
	}
	printf(" ");
	for (i = n & ~0xf; i <= n; ++i) {
		if (!(i & 0x7))
			putchar(' ');
		if (isprint(p[i]))
			putchar(p[i]);
		else
			putchar('.');
	}
	putchar('\n');
}

void
Error(char *str, ...)
{
        va_list ap;
	int e;

	e = errno;

        va_start(ap, str);
        vfprintf(stderr, str, ap);
	va_end(ap);
        fprintf( stderr, "\n" );

	if (e)
		fprintf(stderr, "system error %d: %s\n", e, strerror(e));

        exit(1);
}

char *
getop(int t)
{
	switch (t) {
	    case SSPY_OP_DRV_OPEN:	return "DRV_OPEN";
	    case SSPY_OP_MOD_OPEN:	return "MOD_OPEN";
	    case SSPY_OP_CLOSE:		return "CLOSE";
	    case SSPY_OP_WPUT:		return "WPUT";
	    case SSPY_OP_RPUT:		return "RPUT";
	    default:			return "???";
	}
}

char *
gettype(int t)
{
	switch (t) {
	    case M_DATA:	return "M_DATA";
	    case M_PROTO:	return "M_PROTO";
	    case M_BREAK:	return "M_BREAK";
	    case M_PASSFP:	return "M_PASSFP";
	    case M_EVENT:	return "M_EVENT";
	    case M_SIG:		return "M_SIG";
	    case M_DELAY:	return "M_DELAY";
	    case M_CTL:		return "M_CTL";
	    case M_IOCTL:	return "M_IOCTL";
	    case M_SETOPTS:	return "M_SETOPTS";
	    case M_RSE:		return "M_RSE";
	    case M_IOCACK:	return "M_IOCACK";
	    case M_IOCNAK:	return "M_IOCNAK";
	    case M_PCPROTO:	return "M_PCPROTO";
	    case M_PCSIG:	return "M_PCSIG";
	    case M_READ:	return "M_READ";
	    case M_FLUSH:	return "M_FLUSH";
	    case M_STOP:	return "M_STOP";
	    case M_START:	return "M_START";
	    case M_HANGUP:	return "M_HANGUP";
	    case M_ERROR:	return "M_ERROR";
	    case M_COPYIN:	return "M_COPYIN";
	    case M_COPYOUT:	return "M_COPYOUT";
	    case M_IOCDATA:	return "M_IOCDATA";
	    case M_PCRSE:	return "M_PCRSE";
	    case M_STOPI:	return "M_STOPI";
	    case M_STARTI:	return "M_STARTI";
	    case M_PCEVENT:	return "M_PCEVENT";
	    case M_UNHANGUP:	return "M_UNHANGUP";
	    default: return "???";
	}
}
